;;;;;
;;;;; h l m . s c m		HLL- interpreter in Scheme
;;;;;				interpreter for a Lisp-ish subset

(provide 'hlm)
(require 'lib)		; all the functions
(require 'macrolib)	; all the macros
(require 'envobj)	; syntactic environments
;(require 'pattern)	; Pattern matching


;@ (Module HLL-Interpreter)


;;;;;; Error non-handling
(define (HLL-ERROR msg . rest)
  (apply ERROR (cons (string-append "HLL-: " msg) rest)))


;;;;;;
;;;;;; Multiplexing the environment
;;;;;;

;(define env-struct-definer (new-record-type-definer))
;(define (empty-env))

(define ENV$$mark (make-tag 'env))
(define ENV$$number-field 0)
(define (ENV$$new-field)
   (let ((n ENV$$number-field))
      (set! ENV$$number-field (+ n 1))
      n))
(define ENV$$initializers (deq:new))
(define (ENV:new)
   (let ((do-it (lambda (f) (f))))
      (apply vector (cons ENV$$mark (map do-it (deq:->list ENV$$initializers))))))
(define (ENV$$add-field init)
   (deq:queue! ENV$$initializers init)
   (let ((n (ENV$$new-field)))
      (lambda (x) (vector-ref x n))))


;;;;;;
;;;;;; Main evaluation function
;;;;;;

; Function HI:Val (HLL-interpreter evaluator) has 2 explicit parameters:
;
;  expr = expression to evaluate.
;  env  = environnent in which to evaluate.
;
; The result is (guess what ?)
; the evaluation of the expression in the given environment !
;
; It also has implicit parameters:
;  the meta-environment, which describes recognized object kinds.

; expressions are parsed syntax trees,
;	as given by read or quoted scheme expressions.
; environments are lists
; meta-environment is a list of (predicate action) lists.


(define main-evaluator
  (ref:new
    (lambda (env expr)
      (HLL-ERROR "syntax error:" expr))))

(define (HI:add-handler evaluate-or-continue)
  (ref:set! main-evaluator
     (lambda (env expr)
       (evaluate-or-continue env expr
          (lambda () (main-evaluator env expr))))))

(define (HI:Val expr env)
     ((ref:get main-evaluator) env expr))

;;; And now, let's define the recognized forms...


;;;;;
;;;;; The simplest forms: atoms
;;;;;
(define (atom? x)
   (or (boolean? x) (number? x) (char? x) (string? x)))

(HI:add-handler
  (lambda (env expr cont)
    (if (atom? expr)
       expr
       (cont))))


;------>8------>8------>8------>8------>8------>8------>8------>8------

;;;;;
;;;;; Environments:
;;;;;

(define ENV:bindings (ENV$$add-field syntenv:new))
(define (ENV:get-bindings ENV) (vector-ref ENV ENV:bindings))
(ENV$$add-field syntenv:new))
(*****)

(define ENV:unbound-variable
    (lambda (var)
      (error `("error: unbound variable:" ,var)))) ;@@@

(define (ENV:compare var)
    (lambda (key)
       (eqv? key var)))

(define (ENV:lookup var env eval cont)
  (syntenv:lookup
    (ENV:bindings env)
    (ENV:compare var)
    eval
    cont))

(define (ENV:add-binding env var val)
  (syntenv:... (ENV:bindings env)))

;;;;;
;;;;; Bound Variables:
;;;;;

(define VAR:eval
   (lambda (keva) (cdr keva))

(HI:add-handler
   (lambda (env expr cont)
   (ENV:lookup expr env VAR:eval cont)))


;;;;;
;;;;; Special forms:
;;;;;
(define

;;;;;
;;;;; Lambdas
;;;;;
(define lambda-tag (make-tag))
(define (make-lambda env formals body)
   (list lambda-tag env formals body))
(define (get-fun-env fun) (cadr fun))
(define (get-fun-param-list fun) (caddr fun))
(define (get-fun-body fun) (cadddr fun))

(define (match-formal-parameter env param arg)
   syntenv:cons param arg env)


(define (match-formal-parameters env param-list args)
   (cond
     ((null? param-list)
        (if (null? args) env (HLL-ERROR "too many arguments!" args)))
     ((symbol? param-list)
        (match-formal-parameter env param-list args))
     ((pair? param-list)
        (if (pair? args)
	   (let
	    ((new-env
	      (match-formal-parameter env (car param-list) (car args))))
	    (match-formal-parameters new-env (cdr param-list) (cdr args)))
	   (HLL-ERROR "too few arguments!" args)))))

;;;;;
;;;;; Normal forms:
;;;;;
(define (Apply-Normal fun args)
   (let*
   ((env (get-fun-env fun))
    (param-list (get-fun-param-list fun))
    (thunk (get-fun-body fun))
    (new-env (match-formal-parameters env param-list args)))
   (HI:Val new-env thunk)))

;;;;;
;;;;; Compound forms:
;;;;;
(define (handle-compound env expr)
   (let*
   ((fun (car expr))
    (args (cdr expr))
    (funval (HI:Val env fun)))
   (cond
     ((special-form? funval) (Apply-Special funval args))
     (else (Apply-Normal
     	     funval
             (map (lambda (arg) (Hi:Val env arg) args)))))))

(HI:add-handler
   (lambda (env expr cont)
   (if (pair? expr)
      (handle-compound env expr))))






;;; ---------------------------------------------------------------------------
;;; Transformation du letrec
;;; Le letrec est transforme vers l'application d'une lambda, comme
;;; dans la tranformation d'un let. Les parametres passes ne sont
;;; cependant que des valeurs sans importance (#f). Le corps de la
;;; lambda generee contient les assignations necessaires pour
;;; donner les bonnes valeurs aux variables introduites par le letrec.
;;; Le corps du letrec est ensuite evalue a la fin du corps de la lambda.

(define (transform-letrec expr)
  (if (or (not (= (length expr) 3))
          (not (proper-list? (cadr expr)))
          (not (all (lambda (x) (and (non-empty-list? x) (= (length x) 2)))
                    (cadr expr))))

    (error "`letrec' mal forme")

    (let ((bindings (cadr expr))
          (formals (map1 car (cadr expr)))
          (body    (caddr expr)))
      (cons (list 'lambda formals
                  (list 'begin
                        (cons 'begin
                              (map1 (lambda (x) (cons 'set! x)) (cadr expr)))
                        body))
            (map1 (lambda (x) #f) bindings)))))

(env-def 'letrec transform-letrec syntactic-env)

(define (transform-begin expr)
  (if (not (pair? (cdr expr)))

    (error "`begin' mal termine" expr)

    (if (null? (cddr expr))
      (cadr expr)
      (let ((temp (gensym))) ; variable temporaire
        (list 'let
              (list (list temp (cadr expr)))
              (cons 'begin (cddr expr)))))))

(env-def 'begin transform-begin syntactic-env)




(define-macro% (register-primitive-by-name arg)
  ; we currently drop any typing indication...
  (let ((primitive arg))
     `(register-primitive 'primitive primitive)))

(map-macro
   register-primitive-by-name
   '(null?
     boolean?
     number?
     char?
     string?
     symbol?
     pair?
     list?
     procedure?
     eqv?
     eq?
     equal?
     not
     cons
     car
     cdr
     set-car!
     set-cdr!
     length
     assoc
     member
     <
     >
     <=
     >=
     =
     +
     -
     *
     /
     read
     write
     list
     )); end of registered primitives





;------>8------>8------>8------>8------>8------>8------>8------>8------

;;;
;;; A trivial interface to test it up: read-eval-print loop
;;;

(define (read-eval-print)
    (let ((rep-env  (initial-env)))
    (write 'HLL>)
    (let* ((expr  (read))
      (value (HI:Val expr rep-env)))
    (write value))
    (newline)
    (read-eval-print)))

;;; Let's try it!
;(read-eval-print)
