;;;;
;;;; p a t t e r n . s c m      -- A very simple and stupid reflective
;;;;                               proof checking evaluator for
;;;;                               a pure subset of Scheme
;;;;
;;;; Copyright C 1997 Far Rideau under the GPL
;;;;
;;;;           Author: Far Rideau [rideau@unice.fr]
;;;;

;;;; See README for description

(provide 'pattern)
(require 'lib)
(require 'debug)
(require 'envobj)


(Comment "

Single-minded Patterns follow the following syntax:

variable
(pattern-expression arguments)

A special pattern expression is in full control of its arguments,
but usual patterns match them one by one.

* When inverting a non-surjective pattern,
 it is easy to raise an exception if the pattern is non-invertible
 (assuming that's detectable in a reasonable amount of time!).
* However, when inverting a non-injective pattern,
 what value should be given ?
 = Does the program mean that any matching value should be given,
 and that the program should respect its specification whatever
 value is given ?
 = Should there be backtracking and/or parallel evaluation of
 the possibilities?
 = Should the answer stay unspecified among possible answers?
   Now, if we allow any kind of non-determinism in inverted patterns,
we could as well allow the same kind of non-determinism among patterns.
   It seems to me that the semantically correct way of doing things,
is that a match operation should indeterministically return any matching
term if possible; but because the match pattern need not be linear,
solving indeterminacies cannot be done in a local way, and sometimes,
no computable mechanism can even solve them; thus, reflective escape to
tactics that control this unification are needed. Actually, reflection
allows for several modes of evaluation (see Mercury)...
   Anyway, let us first focus on simple patterns, and pattern language.
")



;;;
;;; Pattern matching
;;;

(Comment

"
simple pattern matching is done by pure functions of the form

(try-match-frame env pat obj) => either nframe #f

where

frame is a frame,
which can be seen as a set of bindings from variable name to value

pat is the pattern to which to match, which is assumed to be of type foo.
obj is the object to match to the pattern

nframe is a new frame in which the object is matched,
#f is returned if the object doesn't match
")


(define (try-match-quote frame qpat obj)
  (if (eqv? ($ quote->obj qpat) obj)
      frame
      #f))

(define (try-match-var frame var obj)
  (frame:try-get frame var
    (lambda (pat) (try-match frame pat obj))
    (lambda () (frame:cons (make-binding var obj) frame))))

(define (try-match-args frame pats objs)
  (cond
   ((and (pair? pats) (pair? objs))
      (let* ((pat1 (car pats))
	     (obj1 (car objs))
	     (pats (cdr pats))
	     (objs (cdr objs))
	     (frame (try-match-expr frame pat1 obj1)))
	(and frame (try-match-args frame pats objs))))
   ((and (null? pats) (null? objs))
      frame)
   (else #f)))

(define (try-match-constr frame pat obj)
  (and (constr? obj)
       (let*
	   ((pat-form (constr:deconstr pat))
	    (pat-cons (car pat-form))
	    (pat-subf (cdr pat-form))
	    (obj-form (constr:deconstr obj))
	    (obj-cons (car obj-form))
	    (obj-subf (cdr obj-form)))
;	 (DBGFMT #t "tMC ~a\n[~a ~a]\n[~a ~a]\n"
;		 frame pat-cons pat-subf obj-cons obj-subf)
	 (and (constr=? pat-cons obj-cons)
	      (eqv? (length pat-subf) (length obj-subf))		;;;[redundant]
	      (try-match-args frame pat-subf obj-subf)))))

; #t is \top and matches everything
; #f is \bottom and matches nothing...
(define (try-match-boolean frame pat obj)
  (if pat frame #f))

(define (try-match-lambda frame pat obj)
  (*****))

(define (try-match-appl frame pat obj)
  (*****))

(define (try-match-special frame pat obj)
  (*****))

(define (try-match-expr frame pat obj)
  (cond
   ((varname? pat)	(try-match-var		frame pat obj))
   ((self-eval? pat)	(try-match-quote	frame ($ make-quote pat) obj))
   ((boolean? pat)	(try-match-boolean	frame pat obj))
   (($ quote? pat)	(try-match-quote	frame pat obj))
   (($ lambda? pat)	(try-match-lambda	frame pat obj))
   (($ appl? pat)	(try-match-appl		frame pat obj))
   (($ special? pat)	(try-match-special	frame pat obj))
;  ((<%? pat)		(try-match-<%		frame pat obj))
   ((constr? pat)	(try-match-constr	frame pat obj))
   (else		#f)))

(define (try-match pat obj)
  (try-match-expr (frame:new) pat obj))

;(traceon try-match)
;(traceon try-match-expr)
;(traceon try-match-var)
;(traceon try-match-quote)
;(traceon try-match-boolean)
;(traceon try-match-lambda)
;(traceon try-match-appl)
;(traceon try-match-special)
;(traceon try-match-constr)
;(traceon try-match-args)

;;;
;;; Unification
;;;

(Comment "
Unification is done by pure functions of the form

(try-unify-foo env lpat rpat y n) => either (y nenv npat) or (n)

where

env is an environment, which can be seen as
either a set of elementary equations between pattern variables
or a graph-representation of lambda-patterns (with variables being nodes).

lpat is the left-hand pattern to unify, which is assumed to be of type foo.
rpat is the right-hand pattern to unify.

nenv is the new environment,
npat is the merged pattern in the new environment...

")

; .....
